# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots['test_types 1'] = {
    'data': {
        '__schema': {
            'types': [
                {
                    'description': None,
                    'fields': [
                        {
                            'description': None,
                            'name': 'version',
                            'type': {
                                'name': 'String'
                            }
                        },
                        {
                            'description': 'The ID of the object',
                            'name': 'node',
                            'type': {
                                'name': 'Node'
                            }
                        },
                        {
                            'description': None,
                            'name': 'webConfig',
                            'type': {
                                'name': 'WebConfig'
                            }
                        },
                        {
                            'description': None,
                            'name': 'allProductTypes',
                            'type': {
                                'name': 'ProductTypeConnection'
                            }
                        },
                        {
                            'description': None,
                            'name': 'productType',
                            'type': {
                                'name': 'ProductType'
                            }
                        },
                        {
                            'description': None,
                            'name': 'allProducts',
                            'type': {
                                'name': 'ProductConnection'
                            }
                        },
                        {
                            'description': None,
                            'name': 'allProductFilePaths',
                            'type': {
                                'name': 'ProductFilePathConnection'
                            }
                        },
                        {
                            'description': None,
                            'name': 'product',
                            'type': {
                                'name': 'Product'
                            }
                        },
                        {
                            'description': None,
                            'name': 'allProductRelationTypes',
                            'type': {
                                'name': 'ProductRelationTypeConnection'
                            }
                        },
                        {
                            'description': None,
                            'name': 'productRelationType',
                            'type': {
                                'name': 'ProductRelationType'
                            }
                        },
                        {
                            'description': None,
                            'name': 'allProductRelations',
                            'type': {
                                'name': 'ProductRelationConnection'
                            }
                        },
                        {
                            'description': None,
                            'name': 'productRelation',
                            'type': {
                                'name': 'ProductRelation'
                            }
                        },
                        {
                            'description': None,
                            'name': 'githubUser',
                            'type': {
                                'name': 'GitHubUser'
                            }
                        },
                        {
                            'description': None,
                            'name': 'oauthAppInfo',
                            'type': {
                                'name': 'OAuthAppInfo'
                            }
                        }
                    ],
                    'inputFields': None,
                    'name': 'Query'
                },
                {
                    'description': 'The `String` scalar type represents textual data, represented as UTF-8 character sequences. The String type is most often used by GraphQL to represent free-form human-readable text.',
                    'fields': None,
                    'inputFields': None,
                    'name': 'String'
                },
                {
                    'description': 'An object with an ID',
                    'fields': [
                        {
                            'description': 'The ID of the object.',
                            'name': 'id',
                            'type': {
                                'name': None
                            }
                        }
                    ],
                    'inputFields': None,
                    'name': 'Node'
                },
                {
                    'description': 'The `ID` scalar type represents a unique identifier, often used to refetch an object or as key for a cache. The ID type appears in a JSON response as a String; however, it is not intended to be human-readable. When expected as an input type, any string (such as `"4"`) or integer (such as `4`) input value will be accepted as an ID.',
                    'fields': None,
                    'inputFields': None,
                    'name': 'ID'
                },
                {
                    'description': 'Web configuration',
                    'fields': [
                        {
                            'description': None,
                            'name': 'configId',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': None,
                            'name': 'headTitle',
                            'type': {
                                'name': 'String'
                            }
                        },
                        {
                            'description': None,
                            'name': 'toolbarTitle',
                            'type': {
                                'name': 'String'
                            }
                        },
                        {
                            'description': None,
                            'name': 'devtoolLoadingstate',
                            'type': {
                                'name': 'Boolean'
                            }
                        },
                        {
                            'description': None,
                            'name': 'productCreationDialog',
                            'type': {
                                'name': 'Boolean'
                            }
                        },
                        {
                            'description': None,
                            'name': 'productUpdateDialog',
                            'type': {
                                'name': 'Boolean'
                            }
                        },
                        {
                            'description': None,
                            'name': 'productDeletionDialog',
                            'type': {
                                'name': 'Boolean'
                            }
                        },
                        {
                            'description': 'The ID of the object.',
                            'name': 'id',
                            'type': {
                                'name': None
                            }
                        }
                    ],
                    'inputFields': None,
                    'name': 'WebConfig'
                },
                {
                    'description': 'The `Boolean` scalar type represents `true` or `false`.',
                    'fields': None,
                    'inputFields': None,
                    'name': 'Boolean'
                },
                {
                    'description': None,
                    'fields': [
                        {
                            'description': 'Pagination data for this connection.',
                            'name': 'pageInfo',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': 'Contains the nodes in this connection.',
                            'name': 'edges',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': None,
                            'name': 'totalCount',
                            'type': {
                                'name': None
                            }
                        }
                    ],
                    'inputFields': None,
                    'name': 'ProductTypeConnection'
                },
                {
                    'description': 'The Relay compliant `PageInfo` type, containing data necessary to paginate this connection.',
                    'fields': [
                        {
                            'description': 'When paginating forwards, are there more items?',
                            'name': 'hasNextPage',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': 'When paginating backwards, are there more items?',
                            'name': 'hasPreviousPage',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': 'When paginating backwards, the cursor to continue.',
                            'name': 'startCursor',
                            'type': {
                                'name': 'String'
                            }
                        },
                        {
                            'description': 'When paginating forwards, the cursor to continue.',
                            'name': 'endCursor',
                            'type': {
                                'name': 'String'
                            }
                        }
                    ],
                    'inputFields': None,
                    'name': 'PageInfo'
                },
                {
                    'description': 'A Relay edge containing a `ProductType` and its cursor.',
                    'fields': [
                        {
                            'description': 'The item at the end of the edge',
                            'name': 'node',
                            'type': {
                                'name': 'ProductType'
                            }
                        },
                        {
                            'description': 'A cursor for use in pagination',
                            'name': 'cursor',
                            'type': {
                                'name': None
                            }
                        }
                    ],
                    'inputFields': None,
                    'name': 'ProductTypeEdge'
                },
                {
                    'description': 'A product type',
                    'fields': [
                        {
                            'description': None,
                            'name': 'typeId',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': None,
                            'name': 'name',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': None,
                            'name': 'order',
                            'type': {
                                'name': 'Int'
                            }
                        },
                        {
                            'description': None,
                            'name': 'indefArticle',
                            'type': {
                                'name': 'String'
                            }
                        },
                        {
                            'description': None,
                            'name': 'singular',
                            'type': {
                                'name': 'String'
                            }
                        },
                        {
                            'description': None,
                            'name': 'plural',
                            'type': {
                                'name': 'String'
                            }
                        },
                        {
                            'description': None,
                            'name': 'icon',
                            'type': {
                                'name': 'String'
                            }
                        },
                        {
                            'description': None,
                            'name': 'products',
                            'type': {
                                'name': 'ProductConnection'
                            }
                        },
                        {
                            'description': 'The ID of the object.',
                            'name': 'id',
                            'type': {
                                'name': None
                            }
                        }
                    ],
                    'inputFields': None,
                    'name': 'ProductType'
                },
                {
                    'description': 'The `Int` scalar type represents non-fractional signed whole numeric values. Int can represent values between -(2^31 - 1) and 2^31 - 1 since represented in JSON as double-precision floating point numbers specifiedby [IEEE 754](http://en.wikipedia.org/wiki/IEEE_floating_point).',
                    'fields': None,
                    'inputFields': None,
                    'name': 'Int'
                },
                {
                    'description': None,
                    'fields': [
                        {
                            'description': 'Pagination data for this connection.',
                            'name': 'pageInfo',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': 'Contains the nodes in this connection.',
                            'name': 'edges',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': None,
                            'name': 'totalCount',
                            'type': {
                                'name': None
                            }
                        }
                    ],
                    'inputFields': None,
                    'name': 'ProductConnection'
                },
                {
                    'description': 'A Relay edge containing a `Product` and its cursor.',
                    'fields': [
                        {
                            'description': 'The item at the end of the edge',
                            'name': 'node',
                            'type': {
                                'name': 'Product'
                            }
                        },
                        {
                            'description': 'A cursor for use in pagination',
                            'name': 'cursor',
                            'type': {
                                'name': None
                            }
                        }
                    ],
                    'inputFields': None,
                    'name': 'ProductEdge'
                },
                {
                    'description': None,
                    'fields': [
                        {
                            'description': None,
                            'name': 'productId',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': None,
                            'name': 'typeId',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': None,
                            'name': 'name',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': None,
                            'name': 'contact',
                            'type': {
                                'name': 'String'
                            }
                        },
                        {
                            'description': None,
                            'name': 'dateProduced',
                            'type': {
                                'name': 'String'
                            }
                        },
                        {
                            'description': None,
                            'name': 'producedBy',
                            'type': {
                                'name': 'String'
                            }
                        },
                        {
                            'description': None,
                            'name': 'datePosted',
                            'type': {
                                'name': 'String'
                            }
                        },
                        {
                            'description': None,
                            'name': 'postedBy',
                            'type': {
                                'name': 'String'
                            }
                        },
                        {
                            'description': None,
                            'name': 'dateUpdated',
                            'type': {
                                'name': 'String'
                            }
                        },
                        {
                            'description': None,
                            'name': 'updatedBy',
                            'type': {
                                'name': 'String'
                            }
                        },
                        {
                            'description': None,
                            'name': 'note',
                            'type': {
                                'name': 'String'
                            }
                        },
                        {
                            'description': None,
                            'name': 'type_',
                            'type': {
                                'name': 'ProductType'
                            }
                        },
                        {
                            'description': None,
                            'name': 'paths',
                            'type': {
                                'name': 'ProductFilePathConnection'
                            }
                        },
                        {
                            'description': None,
                            'name': 'relations',
                            'type': {
                                'name': 'ProductRelationConnection'
                            }
                        },
                        {
                            'description': 'The ID of the object.',
                            'name': 'id',
                            'type': {
                                'name': None
                            }
                        }
                    ],
                    'inputFields': None,
                    'name': 'Product'
                },
                {
                    'description': None,
                    'fields': [
                        {
                            'description': 'Pagination data for this connection.',
                            'name': 'pageInfo',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': 'Contains the nodes in this connection.',
                            'name': 'edges',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': None,
                            'name': 'totalCount',
                            'type': {
                                'name': None
                            }
                        }
                    ],
                    'inputFields': None,
                    'name': 'ProductFilePathConnection'
                },
                {
                    'description': 'A Relay edge containing a `ProductFilePath` and its cursor.',
                    'fields': [
                        {
                            'description': 'The item at the end of the edge',
                            'name': 'node',
                            'type': {
                                'name': 'ProductFilePath'
                            }
                        },
                        {
                            'description': 'A cursor for use in pagination',
                            'name': 'cursor',
                            'type': {
                                'name': None
                            }
                        }
                    ],
                    'inputFields': None,
                    'name': 'ProductFilePathEdge'
                },
                {
                    'description': None,
                    'fields': [
                        {
                            'description': None,
                            'name': 'pathId',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': None,
                            'name': 'path',
                            'type': {
                                'name': 'String'
                            }
                        },
                        {
                            'description': None,
                            'name': 'note',
                            'type': {
                                'name': 'String'
                            }
                        },
                        {
                            'description': None,
                            'name': 'productId',
                            'type': {
                                'name': 'Int'
                            }
                        },
                        {
                            'description': None,
                            'name': 'product',
                            'type': {
                                'name': 'Product'
                            }
                        },
                        {
                            'description': 'The ID of the object.',
                            'name': 'id',
                            'type': {
                                'name': None
                            }
                        }
                    ],
                    'inputFields': None,
                    'name': 'ProductFilePath'
                },
                {
                    'description': 'An enumeration.',
                    'fields': None,
                    'inputFields': None,
                    'name': 'ProductFilePathSortEnum'
                },
                {
                    'description': None,
                    'fields': [
                        {
                            'description': 'Pagination data for this connection.',
                            'name': 'pageInfo',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': 'Contains the nodes in this connection.',
                            'name': 'edges',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': None,
                            'name': 'totalCount',
                            'type': {
                                'name': None
                            }
                        }
                    ],
                    'inputFields': None,
                    'name': 'ProductRelationConnection'
                },
                {
                    'description': 'A Relay edge containing a `ProductRelation` and its cursor.',
                    'fields': [
                        {
                            'description': 'The item at the end of the edge',
                            'name': 'node',
                            'type': {
                                'name': 'ProductRelation'
                            }
                        },
                        {
                            'description': 'A cursor for use in pagination',
                            'name': 'cursor',
                            'type': {
                                'name': None
                            }
                        }
                    ],
                    'inputFields': None,
                    'name': 'ProductRelationEdge'
                },
                {
                    'description': 'A relation from one product to another',
                    'fields': [
                        {
                            'description': None,
                            'name': 'relationId',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': None,
                            'name': 'typeId',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': None,
                            'name': 'selfProductId',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': None,
                            'name': 'otherProductId',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': None,
                            'name': 'reverseRelationId',
                            'type': {
                                'name': 'Int'
                            }
                        },
                        {
                            'description': None,
                            'name': 'type_',
                            'type': {
                                'name': 'ProductRelationType'
                            }
                        },
                        {
                            'description': None,
                            'name': 'self_',
                            'type': {
                                'name': 'Product'
                            }
                        },
                        {
                            'description': None,
                            'name': 'other',
                            'type': {
                                'name': 'Product'
                            }
                        },
                        {
                            'description': None,
                            'name': 'reverse',
                            'type': {
                                'name': 'ProductRelation'
                            }
                        },
                        {
                            'description': 'The ID of the object.',
                            'name': 'id',
                            'type': {
                                'name': None
                            }
                        }
                    ],
                    'inputFields': None,
                    'name': 'ProductRelation'
                },
                {
                    'description': 'A type of relations between products',
                    'fields': [
                        {
                            'description': None,
                            'name': 'typeId',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': None,
                            'name': 'name',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': None,
                            'name': 'reverseTypeId',
                            'type': {
                                'name': 'Int'
                            }
                        },
                        {
                            'description': None,
                            'name': 'indefArticle',
                            'type': {
                                'name': 'String'
                            }
                        },
                        {
                            'description': None,
                            'name': 'singular',
                            'type': {
                                'name': 'String'
                            }
                        },
                        {
                            'description': None,
                            'name': 'plural',
                            'type': {
                                'name': 'String'
                            }
                        },
                        {
                            'description': None,
                            'name': 'reverse',
                            'type': {
                                'name': 'ProductRelationType'
                            }
                        },
                        {
                            'description': None,
                            'name': 'relations',
                            'type': {
                                'name': 'ProductRelationConnection'
                            }
                        },
                        {
                            'description': 'The ID of the object.',
                            'name': 'id',
                            'type': {
                                'name': None
                            }
                        }
                    ],
                    'inputFields': None,
                    'name': 'ProductRelationType'
                },
                {
                    'description': 'An enumeration.',
                    'fields': None,
                    'inputFields': None,
                    'name': 'ProductRelationSortEnum'
                },
                {
                    'description': None,
                    'fields': None,
                    'inputFields': [
                        {
                            'defaultValue': None,
                            'description': 'Exact match.',
                            'name': 'typeId'
                        },
                        {
                            'defaultValue': None,
                            'description': 'Conjunction of filters joined by ``AND``.',
                            'name': 'and'
                        },
                        {
                            'defaultValue': None,
                            'description': 'Conjunction of filters joined by ``OR``.',
                            'name': 'or'
                        },
                        {
                            'defaultValue': None,
                            'description': 'Negation of filters.',
                            'name': 'not'
                        },
                        {
                            'defaultValue': None,
                            'description': None,
                            'name': 'typeName'
                        }
                    ],
                    'name': 'ProductFilter'
                },
                {
                    'description': 'An enumeration.',
                    'fields': None,
                    'inputFields': None,
                    'name': 'ProductSortEnum'
                },
                {
                    'description': None,
                    'fields': None,
                    'inputFields': [
                        {
                            'defaultValue': None,
                            'description': 'Conjunction of filters joined by ``AND``.',
                            'name': 'and'
                        },
                        {
                            'defaultValue': None,
                            'description': 'Conjunction of filters joined by ``OR``.',
                            'name': 'or'
                        },
                        {
                            'defaultValue': None,
                            'description': 'Negation of filters.',
                            'name': 'not'
                        }
                    ],
                    'name': 'ProductTypeFilter'
                },
                {
                    'description': 'An enumeration.',
                    'fields': None,
                    'inputFields': None,
                    'name': 'ProductTypeSortEnum'
                },
                {
                    'description': None,
                    'fields': [
                        {
                            'description': 'Pagination data for this connection.',
                            'name': 'pageInfo',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': 'Contains the nodes in this connection.',
                            'name': 'edges',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': None,
                            'name': 'totalCount',
                            'type': {
                                'name': None
                            }
                        }
                    ],
                    'inputFields': None,
                    'name': 'ProductRelationTypeConnection'
                },
                {
                    'description': 'A Relay edge containing a `ProductRelationType` and its cursor.',
                    'fields': [
                        {
                            'description': 'The item at the end of the edge',
                            'name': 'node',
                            'type': {
                                'name': 'ProductRelationType'
                            }
                        },
                        {
                            'description': 'A cursor for use in pagination',
                            'name': 'cursor',
                            'type': {
                                'name': None
                            }
                        }
                    ],
                    'inputFields': None,
                    'name': 'ProductRelationTypeEdge'
                },
                {
                    'description': 'An enumeration.',
                    'fields': None,
                    'inputFields': None,
                    'name': 'ProductRelationTypeSortEnum'
                },
                {
                    'description': None,
                    'fields': [
                        {
                            'description': None,
                            'name': 'login',
                            'type': {
                                'name': 'String'
                            }
                        },
                        {
                            'description': None,
                            'name': 'name',
                            'type': {
                                'name': 'String'
                            }
                        },
                        {
                            'description': None,
                            'name': 'avatarUrl',
                            'type': {
                                'name': 'String'
                            }
                        }
                    ],
                    'inputFields': None,
                    'name': 'GitHubUser'
                },
                {
                    'description': None,
                    'fields': [
                        {
                            'description': None,
                            'name': 'clientId',
                            'type': {
                                'name': 'String'
                            }
                        },
                        {
                            'description': None,
                            'name': 'authorizeUrl',
                            'type': {
                                'name': 'String'
                            }
                        },
                        {
                            'description': None,
                            'name': 'tokenUrl',
                            'type': {
                                'name': 'String'
                            }
                        },
                        {
                            'description': None,
                            'name': 'redirectUri',
                            'type': {
                                'name': 'String'
                            }
                        }
                    ],
                    'inputFields': None,
                    'name': 'OAuthAppInfo'
                },
                {
                    'description': None,
                    'fields': [
                        {
                            'description': 'Create a product',
                            'name': 'createProduct',
                            'type': {
                                'name': 'CreateProduct'
                            }
                        },
                        {
                            'description': '''Update a product. Note: This is to update the DB entry about a product. If the
product itself has been updated, a new entry should be added by
createProduct()''',
                            'name': 'updateProduct',
                            'type': {
                                'name': 'UpdateProduct'
                            }
                        },
                        {
                            'description': 'Delete a product',
                            'name': 'deleteProduct',
                            'type': {
                                'name': 'DeleteProduct'
                            }
                        },
                        {
                            'description': None,
                            'name': 'createProductFilePath',
                            'type': {
                                'name': 'CreateProductFilePath'
                            }
                        },
                        {
                            'description': None,
                            'name': 'updateProductFilePath',
                            'type': {
                                'name': 'UpdateProductFilePath'
                            }
                        },
                        {
                            'description': None,
                            'name': 'deleteProductFilePath',
                            'type': {
                                'name': 'DeleteProductFilePath'
                            }
                        },
                        {
                            'description': 'Create a product type',
                            'name': 'createProductType',
                            'type': {
                                'name': 'CreateProductType'
                            }
                        },
                        {
                            'description': 'Update a product type',
                            'name': 'updateProductType',
                            'type': {
                                'name': 'UpdateProductType'
                            }
                        },
                        {
                            'description': 'Delete a product type',
                            'name': 'deleteProductType',
                            'type': {
                                'name': 'DeleteProductType'
                            }
                        },
                        {
                            'description': 'Create a pair of product relation types',
                            'name': 'createProductRelationTypes',
                            'type': {
                                'name': 'CreateProductRelationTypes'
                            }
                        },
                        {
                            'description': 'Update a product relation type',
                            'name': 'updateProductRelationType',
                            'type': {
                                'name': 'UpdateProductRelationType'
                            }
                        },
                        {
                            'description': 'Delete a pair of product relation types',
                            'name': 'deleteProductRelationTypes',
                            'type': {
                                'name': 'DeleteProductRelationTypes'
                            }
                        },
                        {
                            'description': '''Add relations between two products. The arguments only specify the relation
from one product to the other. The reverse relation will be also added.''',
                            'name': 'createProductRelation',
                            'type': {
                                'name': 'CreateProductRelation'
                            }
                        },
                        {
                            'description': '''Remove relations from two products.

    ''',
                            'name': 'deleteProductRelation',
                            'type': {
                                'name': 'DeleteProductRelation'
                            }
                        },
                        {
                            'description': None,
                            'name': 'githubAuth',
                            'type': {
                                'name': 'GitHubAuth'
                            }
                        },
                        {
                            'description': None,
                            'name': 'storeAdminAppToken',
                            'type': {
                                'name': 'StoreAdminAppToken'
                            }
                        }
                    ],
                    'inputFields': None,
                    'name': 'Mutation'
                },
                {
                    'description': 'Create a product',
                    'fields': [
                        {
                            'description': None,
                            'name': 'ok',
                            'type': {
                                'name': 'Boolean'
                            }
                        },
                        {
                            'description': None,
                            'name': 'product',
                            'type': {
                                'name': 'Product'
                            }
                        }
                    ],
                    'inputFields': None,
                    'name': 'CreateProduct'
                },
                {
                    'description': 'Input to createProduct()',
                    'fields': None,
                    'inputFields': [
                        {
                            'defaultValue': None,
                            'description': 'A person or group that can be contacted for questions or issues about the product.',
                            'name': 'contact'
                        },
                        {
                            'defaultValue': None,
                            'description': 'Note about the product in MarkDown.',
                            'name': 'note'
                        },
                        {
                            'defaultValue': None,
                            'description': 'Paths to the products. e.g., nersc:/go/to/my/product_v3',
                            'name': 'paths'
                        },
                        {
                            'defaultValue': None,
                            'description': 'Relations to other products',
                            'name': 'relations'
                        },
                        {
                            'defaultValue': None,
                            'description': 'The product type ID',
                            'name': 'typeId'
                        },
                        {
                            'defaultValue': None,
                            'description': 'The name of the product',
                            'name': 'name'
                        },
                        {
                            'defaultValue': None,
                            'description': 'The date on which the product was produced',
                            'name': 'dateProduced'
                        },
                        {
                            'defaultValue': None,
                            'description': 'The person or group that produced the product',
                            'name': 'producedBy'
                        },
                        {
                            'defaultValue': None,
                            'description': 'The person who entered the DB entry.',
                            'name': 'postedBy'
                        }
                    ],
                    'name': 'CreateProductInput'
                },
                {
                    'description': 'A relation to another product',
                    'fields': None,
                    'inputFields': [
                        {
                            'defaultValue': None,
                            'description': 'The product ID of the other product',
                            'name': 'productId'
                        },
                        {
                            'defaultValue': None,
                            'description': 'The relation type ID',
                            'name': 'typeId'
                        }
                    ],
                    'name': 'RelationInputFields'
                },
                {
                    'description': '''The `Date` scalar type represents a Date
value as specified by
[iso8601](https://en.wikipedia.org/wiki/ISO_8601).''',
                    'fields': None,
                    'inputFields': None,
                    'name': 'Date'
                },
                {
                    'description': '''Update a product. Note: This is to update the DB entry about a product. If the
product itself has been updated, a new entry should be added by
createProduct()''',
                    'fields': [
                        {
                            'description': None,
                            'name': 'ok',
                            'type': {
                                'name': 'Boolean'
                            }
                        },
                        {
                            'description': None,
                            'name': 'product',
                            'type': {
                                'name': 'Product'
                            }
                        }
                    ],
                    'inputFields': None,
                    'name': 'UpdateProduct'
                },
                {
                    'description': 'Input to updateProduct()',
                    'fields': None,
                    'inputFields': [
                        {
                            'defaultValue': None,
                            'description': 'A person or group that can be contacted for questions or issues about the product.',
                            'name': 'contact'
                        },
                        {
                            'defaultValue': None,
                            'description': 'Note about the product in MarkDown.',
                            'name': 'note'
                        },
                        {
                            'defaultValue': None,
                            'description': 'Paths to the products. e.g., nersc:/go/to/my/product_v3',
                            'name': 'paths'
                        },
                        {
                            'defaultValue': None,
                            'description': 'Relations to other products',
                            'name': 'relations'
                        },
                        {
                            'defaultValue': None,
                            'description': 'The person who updated the DB entry.',
                            'name': 'updatedBy'
                        }
                    ],
                    'name': 'UpdateProductInput'
                },
                {
                    'description': 'Delete a product',
                    'fields': [
                        {
                            'description': None,
                            'name': 'ok',
                            'type': {
                                'name': 'Boolean'
                            }
                        }
                    ],
                    'inputFields': None,
                    'name': 'DeleteProduct'
                },
                {
                    'description': None,
                    'fields': [
                        {
                            'description': None,
                            'name': 'ok',
                            'type': {
                                'name': 'Boolean'
                            }
                        },
                        {
                            'description': None,
                            'name': 'productFilePath',
                            'type': {
                                'name': 'ProductFilePath'
                            }
                        }
                    ],
                    'inputFields': None,
                    'name': 'CreateProductFilePath'
                },
                {
                    'description': None,
                    'fields': None,
                    'inputFields': [
                        {
                            'defaultValue': None,
                            'description': None,
                            'name': 'path'
                        },
                        {
                            'defaultValue': None,
                            'description': None,
                            'name': 'note'
                        },
                        {
                            'defaultValue': None,
                            'description': None,
                            'name': 'productId'
                        }
                    ],
                    'name': 'CreateProductFilePathInput'
                },
                {
                    'description': None,
                    'fields': [
                        {
                            'description': None,
                            'name': 'ok',
                            'type': {
                                'name': 'Boolean'
                            }
                        },
                        {
                            'description': None,
                            'name': 'productFilePath',
                            'type': {
                                'name': 'ProductFilePath'
                            }
                        }
                    ],
                    'inputFields': None,
                    'name': 'UpdateProductFilePath'
                },
                {
                    'description': None,
                    'fields': None,
                    'inputFields': [
                        {
                            'defaultValue': None,
                            'description': None,
                            'name': 'path'
                        },
                        {
                            'defaultValue': None,
                            'description': None,
                            'name': 'note'
                        }
                    ],
                    'name': 'UpdateProductFilePathInput'
                },
                {
                    'description': None,
                    'fields': [
                        {
                            'description': None,
                            'name': 'ok',
                            'type': {
                                'name': 'Boolean'
                            }
                        }
                    ],
                    'inputFields': None,
                    'name': 'DeleteProductFilePath'
                },
                {
                    'description': 'Create a product type',
                    'fields': [
                        {
                            'description': None,
                            'name': 'ok',
                            'type': {
                                'name': 'Boolean'
                            }
                        },
                        {
                            'description': None,
                            'name': 'productType',
                            'type': {
                                'name': 'ProductType'
                            }
                        }
                    ],
                    'inputFields': None,
                    'name': 'CreateProductType'
                },
                {
                    'description': 'Input to createProductType()',
                    'fields': None,
                    'inputFields': [
                        {
                            'defaultValue': None,
                            'description': 'The order in which the type is displayed, for example, in navigation bars.',
                            'name': 'order'
                        },
                        {
                            'defaultValue': None,
                            'description': 'The indefinite article placed before the singular noun "i.e., "a" or "an". ',
                            'name': 'indefArticle'
                        },
                        {
                            'defaultValue': None,
                            'description': 'The singular noun, the product type name in singular.',
                            'name': 'singular'
                        },
                        {
                            'defaultValue': None,
                            'description': 'The plural noun, the product type name in plural.',
                            'name': 'plural'
                        },
                        {
                            'defaultValue': None,
                            'description': 'A name of the icon from https://materialdesignicons.com/',
                            'name': 'icon'
                        },
                        {
                            'defaultValue': None,
                            'description': 'The name of the product type',
                            'name': 'name'
                        }
                    ],
                    'name': 'CreateProductTypeInput'
                },
                {
                    'description': 'Update a product type',
                    'fields': [
                        {
                            'description': None,
                            'name': 'ok',
                            'type': {
                                'name': 'Boolean'
                            }
                        },
                        {
                            'description': None,
                            'name': 'productType',
                            'type': {
                                'name': 'ProductType'
                            }
                        }
                    ],
                    'inputFields': None,
                    'name': 'UpdateProductType'
                },
                {
                    'description': 'Input to updateProductType()',
                    'fields': None,
                    'inputFields': [
                        {
                            'defaultValue': None,
                            'description': 'The order in which the type is displayed, for example, in navigation bars.',
                            'name': 'order'
                        },
                        {
                            'defaultValue': None,
                            'description': 'The indefinite article placed before the singular noun "i.e., "a" or "an". ',
                            'name': 'indefArticle'
                        },
                        {
                            'defaultValue': None,
                            'description': 'The singular noun, the product type name in singular.',
                            'name': 'singular'
                        },
                        {
                            'defaultValue': None,
                            'description': 'The plural noun, the product type name in plural.',
                            'name': 'plural'
                        },
                        {
                            'defaultValue': None,
                            'description': 'A name of the icon from https://materialdesignicons.com/',
                            'name': 'icon'
                        }
                    ],
                    'name': 'UpdateProductTypeInput'
                },
                {
                    'description': 'Delete a product type',
                    'fields': [
                        {
                            'description': None,
                            'name': 'ok',
                            'type': {
                                'name': 'Boolean'
                            }
                        }
                    ],
                    'inputFields': None,
                    'name': 'DeleteProductType'
                },
                {
                    'description': 'Create a pair of product relation types',
                    'fields': [
                        {
                            'description': None,
                            'name': 'ok',
                            'type': {
                                'name': 'Boolean'
                            }
                        },
                        {
                            'description': None,
                            'name': 'productRelationType',
                            'type': {
                                'name': 'ProductRelationType'
                            }
                        }
                    ],
                    'inputFields': None,
                    'name': 'CreateProductRelationTypes'
                },
                {
                    'description': 'An input to createProductRelationTypes()',
                    'fields': None,
                    'inputFields': [
                        {
                            'defaultValue': None,
                            'description': 'The indefinite article placed before the singular noun "i.e., "a" or "an". ',
                            'name': 'indefArticle'
                        },
                        {
                            'defaultValue': None,
                            'description': 'The singular noun, the relation type name in singular.',
                            'name': 'singular'
                        },
                        {
                            'defaultValue': None,
                            'description': 'The plural noun, the relation type name in plural.',
                            'name': 'plural'
                        },
                        {
                            'defaultValue': None,
                            'description': 'The name of the relation type',
                            'name': 'name'
                        }
                    ],
                    'name': 'CreateProductRelationTypeInput'
                },
                {
                    'description': 'Update a product relation type',
                    'fields': [
                        {
                            'description': None,
                            'name': 'ok',
                            'type': {
                                'name': 'Boolean'
                            }
                        },
                        {
                            'description': None,
                            'name': 'productRelationType',
                            'type': {
                                'name': 'ProductRelationType'
                            }
                        }
                    ],
                    'inputFields': None,
                    'name': 'UpdateProductRelationType'
                },
                {
                    'description': 'An input to updateProductRelationType()',
                    'fields': None,
                    'inputFields': [
                        {
                            'defaultValue': None,
                            'description': 'The indefinite article placed before the singular noun "i.e., "a" or "an". ',
                            'name': 'indefArticle'
                        },
                        {
                            'defaultValue': None,
                            'description': 'The singular noun, the relation type name in singular.',
                            'name': 'singular'
                        },
                        {
                            'defaultValue': None,
                            'description': 'The plural noun, the relation type name in plural.',
                            'name': 'plural'
                        }
                    ],
                    'name': 'UpdateProductRelationTypeInput'
                },
                {
                    'description': 'Delete a pair of product relation types',
                    'fields': [
                        {
                            'description': None,
                            'name': 'ok',
                            'type': {
                                'name': 'Boolean'
                            }
                        }
                    ],
                    'inputFields': None,
                    'name': 'DeleteProductRelationTypes'
                },
                {
                    'description': '''Add relations between two products. The arguments only specify the relation
from one product to the other. The reverse relation will be also added.''',
                    'fields': [
                        {
                            'description': None,
                            'name': 'ok',
                            'type': {
                                'name': 'Boolean'
                            }
                        },
                        {
                            'description': None,
                            'name': 'productRelation',
                            'type': {
                                'name': 'ProductRelation'
                            }
                        }
                    ],
                    'inputFields': None,
                    'name': 'CreateProductRelation'
                },
                {
                    'description': 'An input to createProductRelation()',
                    'fields': None,
                    'inputFields': [
                        {
                            'defaultValue': None,
                            'description': 'The typeId of the product relation type of the relation from "self" to the "other"',
                            'name': 'typeId'
                        },
                        {
                            'defaultValue': None,
                            'description': 'The productId of the self product',
                            'name': 'selfProductId'
                        },
                        {
                            'defaultValue': None,
                            'description': 'The productId of the other product',
                            'name': 'otherProductId'
                        }
                    ],
                    'name': 'CreateProductRelationInput'
                },
                {
                    'description': '''Remove relations from two products.

    ''',
                    'fields': [
                        {
                            'description': None,
                            'name': 'ok',
                            'type': {
                                'name': 'Boolean'
                            }
                        }
                    ],
                    'inputFields': None,
                    'name': 'DeleteProductRelation'
                },
                {
                    'description': None,
                    'fields': [
                        {
                            'description': None,
                            'name': 'authPayload',
                            'type': {
                                'name': 'AuthPayload'
                            }
                        }
                    ],
                    'inputFields': None,
                    'name': 'GitHubAuth'
                },
                {
                    'description': None,
                    'fields': [
                        {
                            'description': None,
                            'name': 'token',
                            'type': {
                                'name': 'String'
                            }
                        }
                    ],
                    'inputFields': None,
                    'name': 'AuthPayload'
                },
                {
                    'description': None,
                    'fields': [
                        {
                            'description': None,
                            'name': 'ok',
                            'type': {
                                'name': 'Boolean'
                            }
                        }
                    ],
                    'inputFields': None,
                    'name': 'StoreAdminAppToken'
                },
                {
                    'description': 'A GraphQL Schema defines the capabilities of a GraphQL server. It exposes all available types and directives on the server, as well as the entry points for query, mutation and subscription operations.',
                    'fields': [
                        {
                            'description': 'A list of all types supported by this server.',
                            'name': 'types',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': 'The type that query operations will be rooted at.',
                            'name': 'queryType',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': 'If this server supports mutation, the type that mutation operations will be rooted at.',
                            'name': 'mutationType',
                            'type': {
                                'name': '__Type'
                            }
                        },
                        {
                            'description': 'If this server support subscription, the type that subscription operations will be rooted at.',
                            'name': 'subscriptionType',
                            'type': {
                                'name': '__Type'
                            }
                        },
                        {
                            'description': 'A list of all directives supported by this server.',
                            'name': 'directives',
                            'type': {
                                'name': None
                            }
                        }
                    ],
                    'inputFields': None,
                    'name': '__Schema'
                },
                {
                    'description': '''The fundamental unit of any GraphQL Schema is the type. There are many kinds of types in GraphQL as represented by the `__TypeKind` enum.

Depending on the kind of a type, certain fields describe information about that type. Scalar types provide no information beyond a name and description, while Enum types provide their values. Object and Interface types provide the fields they describe. Abstract types, Union and Interface, provide the Object types possible at runtime. List and NonNull types compose other types.''',
                    'fields': [
                        {
                            'description': None,
                            'name': 'kind',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': None,
                            'name': 'name',
                            'type': {
                                'name': 'String'
                            }
                        },
                        {
                            'description': None,
                            'name': 'description',
                            'type': {
                                'name': 'String'
                            }
                        },
                        {
                            'description': None,
                            'name': 'fields',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': None,
                            'name': 'interfaces',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': None,
                            'name': 'possibleTypes',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': None,
                            'name': 'enumValues',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': None,
                            'name': 'inputFields',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': None,
                            'name': 'ofType',
                            'type': {
                                'name': '__Type'
                            }
                        }
                    ],
                    'inputFields': None,
                    'name': '__Type'
                },
                {
                    'description': 'An enum describing what kind of type a given `__Type` is',
                    'fields': None,
                    'inputFields': None,
                    'name': '__TypeKind'
                },
                {
                    'description': 'Object and Interface types are described by a list of Fields, each of which has a name, potentially a list of arguments, and a return type.',
                    'fields': [
                        {
                            'description': None,
                            'name': 'name',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': None,
                            'name': 'description',
                            'type': {
                                'name': 'String'
                            }
                        },
                        {
                            'description': None,
                            'name': 'args',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': None,
                            'name': 'type',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': None,
                            'name': 'isDeprecated',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': None,
                            'name': 'deprecationReason',
                            'type': {
                                'name': 'String'
                            }
                        }
                    ],
                    'inputFields': None,
                    'name': '__Field'
                },
                {
                    'description': 'Arguments provided to Fields or Directives and the input fields of an InputObject are represented as Input Values which describe their type and optionally a default value.',
                    'fields': [
                        {
                            'description': None,
                            'name': 'name',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': None,
                            'name': 'description',
                            'type': {
                                'name': 'String'
                            }
                        },
                        {
                            'description': None,
                            'name': 'type',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': None,
                            'name': 'defaultValue',
                            'type': {
                                'name': 'String'
                            }
                        }
                    ],
                    'inputFields': None,
                    'name': '__InputValue'
                },
                {
                    'description': 'One possible value for a given Enum. Enum values are unique values, not a placeholder for a string or numeric value. However an Enum value is returned in a JSON response as a string.',
                    'fields': [
                        {
                            'description': None,
                            'name': 'name',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': None,
                            'name': 'description',
                            'type': {
                                'name': 'String'
                            }
                        },
                        {
                            'description': None,
                            'name': 'isDeprecated',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': None,
                            'name': 'deprecationReason',
                            'type': {
                                'name': 'String'
                            }
                        }
                    ],
                    'inputFields': None,
                    'name': '__EnumValue'
                },
                {
                    'description': '''A Directive provides a way to describe alternate runtime execution and type validation behavior in a GraphQL document.

In some cases, you need to provide options to alter GraphQL's execution behavior in ways field arguments will not suffice, such as conditionally including or skipping a field. Directives provide this by describing additional information to the executor.''',
                    'fields': [
                        {
                            'description': None,
                            'name': 'name',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': None,
                            'name': 'description',
                            'type': {
                                'name': 'String'
                            }
                        },
                        {
                            'description': None,
                            'name': 'locations',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': None,
                            'name': 'args',
                            'type': {
                                'name': None
                            }
                        }
                    ],
                    'inputFields': None,
                    'name': '__Directive'
                },
                {
                    'description': 'A Directive can be adjacent to many parts of the GraphQL language, a __DirectiveLocation describes one such possible adjacencies.',
                    'fields': None,
                    'inputFields': None,
                    'name': '__DirectiveLocation'
                }
            ]
        }
    }
}



import pytest

import datetime

from acondbs import create_app
from acondbs.db.ops import define_tables

from acondbs.db.sa import sa
from acondbs.models import (
    ProductType,
    Product,
    ProductFilePath,
    ProductRelationType,
    ProductRelation
    )

##__________________________________________________________________||
@pytest.fixture
def app_empty():
    database_uri ="sqlite:///:memory:"
    y = create_app(SQLALCHEMY_DATABASE_URI=database_uri)
    with y.app_context():
        define_tables()
    yield y

@pytest.fixture
def app(app_empty):

    y = app_empty

    # map1 -> beam1
    #            |
    #            +---> beam2
    #
    # map2
    # map3

    # create relation types
    Parent = ProductRelationType(type_id=1, name='parent')
    Parent.reverse = ProductRelationType(type_id=2, name='child')

    # create product types
    Map = ProductType(type_id=1, name='map')
    Beam = ProductType(type_id=2, name='beam')

    # create products
    map1 = Product(product_id=1, name="map1", date_produced=datetime.date(2020, 2, 1), type_=Map)
    map2 = Product(product_id=2, name="map2", date_produced=datetime.date(2020, 2, 10), type_=Map)
    map3 = Product(product_id=3, name="map3", date_produced=datetime.date(2020, 3, 19), type_=Map)
    beam1 = Product(product_id=4, name="beam1", date_produced=datetime.date(2020, 2, 5), type_=Beam)
    beam2 = Product(product_id=5, name="beam2", date_produced=datetime.date(2020, 3, 4), type_=Beam)

    # create relations
    relation1 = ProductRelation(relation_id=2, type_=Parent, self_=beam1, other=map1)
    relation2 = ProductRelation(relation_id=4, type_=Parent, self_=beam2, other=beam1)

    with y.app_context():
        sa.session.add(Map)
        sa.session.add(Beam)
        sa.session.commit()
    yield y

##__________________________________________________________________||



"""define filters

Define filters by using graphene-sqlalchemy-filter:
https://github.com/art1415926535/graphene-sqlalchemy-filter

example:
https://github.com/art1415926535/graphene-sqlalchemy-filter/blob/1.10.2/examples/clients_and_records/filters.py

"""
import graphene
from graphene_sqlalchemy_filter import FilterableConnectionField, FilterSet

from ..models import Product as ProductModel
from ..models import ProductType as ProductTypeModel

##__________________________________________________________________||
class ProductFilter(FilterSet):
   type_name = graphene.String()

   class Meta:
       model = ProductModel
       fields = {
           'type_id': ['eq', ],
       }

   @staticmethod
   def type_name_filter(info, query, value):
       # "Filters that require join":
       # https://github.com/art1415926535/graphene-sqlalchemy-filter/tree/1.10.2#filters-that-require-join
       query = query.join(ProductTypeModel)
       filter_ = ProductTypeModel.name == value
       return query, filter_

class ProductTypeFilter(FilterSet):
   class Meta:
       model = ProductTypeModel
       fields = { }

##__________________________________________________________________||
class PFilterableConnectionField(FilterableConnectionField):
    filters = {
        ProductModel: ProductFilter(),
        ProductTypeModel: ProductTypeFilter()
    }

##__________________________________________________________________||




import pytest
import unittest.mock as mock

from acondbs.misc import githubauth

##__________________________________________________________________||
@pytest.fixture(autouse=True)
def mock_requests(monkeypatch):
    y = mock.Mock()
    monkeypatch.setattr("acondbs.misc.githubauth.requests", y)
    yield y


##__________________________________________________________________||
def test_success(app, mock_requests):

    code = 'code-xyz'

    r = {'access_token': 'token-xxx', 'token_type': 'bearer', 'scope': 'user'}
    mock_requests.post().json.return_value = r
    with app.app_context():
        token = githubauth.get_token(code)

    assert 'token-xxx' == token

##__________________________________________________________________||
def test_error(app, mock_requests):

    code = 'code-xyz'

    r = {
        'error': 'bad_verification_code',
        'error_description': 'The code passed is incorrect or expired.',
        'error_uri': 'https://docs.github.com/apps/managing-oauth-apps/troubleshooting-oauth-app-access-token-request-errors/#bad-verification-code'
    }
    mock_requests.post().json.return_value = r
    with app.app_context():
        token = githubauth.get_token(code)

    assert token is None

##__________________________________________________________________||



import pytest
import textwrap

from ..funcs import assert_query

from ..gql import (
    FRAGMENT_PRODUCT_CONNECTION_SHALLOW,
    FRAGMENT_PRODUCT_CONNECTION_DEEP
    )

##__________________________________________________________________||
params = [
    pytest.param(
        [textwrap.dedent('''
          {
            allProductFilePaths {
              edges {
                node {
                  pathId
                  path
                  note
                }
              }
            }
          }
        '''),],
        {},
        id='all'
    ),
    pytest.param(
        [textwrap.dedent('''
          {
            allProductFilePaths {
              totalCount
            }
          }
        '''),],
        {},
        id='total-count'
    ),
]

##__________________________________________________________________||
@pytest.mark.parametrize('args, kwags', params)
def test_schema(app, snapshot, args, kwags):
    assert_query(app, snapshot, [args, kwags])

##__________________________________________________________________||



import pytest

from acondbs import create_app
from acondbs.db.ops import define_tables

from acondbs.db.sa import sa
from acondbs.models import ProductType, Product, ProductRelation, ProductRelationType

##__________________________________________________________________||
@pytest.fixture
def app_empty():
    database_uri ="sqlite:///:memory:"
    y = create_app(SQLALCHEMY_DATABASE_URI=database_uri)
    with y.app_context():
        define_tables()
    yield y

@pytest.fixture
def app(app_empty):

    y = app_empty

    #                              +--------+
    #               --(child)-->   |        |
    #                    |         | child1 |
    #  +---------+  <-(parent)--   |        |
    #  |         |                 +--------+
    #  | parent1 |
    #  |         |                 +--------+
    #  +---------+  --(child)-->   |        |
    #                    |         | child2 |
    #               <-(parent)--   |        |
    #                              +--------+

    type_ = ProductType(name='robot')
    parent1 = Product(name="parent1", type_=type_)
    child1 = Product(name="child1", type_=type_)
    child2 = Product(name="child2", type_=type_)

    relation_type_parent = ProductRelationType(name='parent')
    relation_type_child = ProductRelationType(name='child')
    relation_type_parent.reverse = relation_type_child

    # parent1 --(child)--> child1
    relation_parent1_to_child1 = ProductRelation()
    relation_parent1_to_child1.type_ = relation_type_child
    relation_parent1_to_child1.self_ = parent1
    relation_parent1_to_child1.other = child1

    ## the reverse relation child1 --(parent)--> parent1 will be
    ## automatically set

    # parent1 --(child)--> child2
    relation_parent1_to_child2 = ProductRelation()
    relation_parent1_to_child2.type_ = relation_type_child
    relation_parent1_to_child2.self_ = parent1
    relation_parent1_to_child2.other = child2

    ## the reverse relation child2 --(parent)--> parent1 will be
    ## automatically set

    # commit
    with y.app_context():
        sa.session.add(parent1)
        sa.session.commit()
    yield y

# __________________________________________________________________||



import pytest

from acondbs import create_app
from acondbs.db.ops import define_tables

##__________________________________________________________________||
@pytest.fixture
def app_empty():
    database_uri ="sqlite:///:memory:"
    y = create_app(SQLALCHEMY_DATABASE_URI=database_uri)
    with y.app_context():
        define_tables()
    yield y

# __________________________________________________________________||



from graphene.test import Client

from acondbs.schema.schema import create_schema

##__________________________________________________________________||
def assert_query(app, snapshot, query, error=False):

    # create test client
    # https://docs.graphene-python.org/en/latest/testing/
    schema = create_schema()
    client = Client(schema)

    # arguments to client.execute()
    args, kwags = query

    if 'context_value' not in kwags:
        # provide an empty dict to prevent errors
        # in graphene_sqlalchemy_filter.
        kwags['context_value'] = {}

        # The detail is described in the commit message:
        #   https://github.com/simonsobs/acondbs/commit/0c65d5719129c4940eea0763ae7b9cc1bcddfe64
        # A related issue in graphene:
        #   https://github.com/graphql-python/graphene/issues/591

    # execute the query
    with app.app_context():
        result = client.execute(*args, **kwags)

    # assert errors
    if error:
        assert 'errors' in result
        return

    assert 'errors' not in result

    # snapshot test
    #   https://github.com/syrusakbary/snapshottest/
    snapshot.assert_match(result)

def assert_mutation(app, snapshot, mutation, query, mock_request_backup_db, success=True):
    assert_query(app, snapshot, mutation, error=not success)
    assert_query(app, snapshot, query)

##__________________________________________________________________||



from ..db.sa import sa

##__________________________________________________________________||
class ProductFilePath(sa.Model):
    __tablename__ = 'product_file_paths'
    path_id = sa.Column(sa.Integer(), primary_key=True)
    path = sa.Column(sa.Text())
    note = sa.Column(sa.Text())
    product_id = sa.Column(sa.ForeignKey('products.product_id'))
    product = sa.relationship(
        "Product",
        backref=sa.backref("paths", cascade="all, delete-orphan"))

##__________________________________________________________________||



import pytest

from ..funcs import assert_mutation

from ..gql import (
    FRAGMENT_PRODUCT_RELATION_TYPE_CONNECTION,
    UPDATE_PRODUCT_RELATION_TYPE
    )

##__________________________________________________________________||
QEURY = '''
{
  allProductRelationTypes {
   ...fragmentProductRelationTypeConnection
  }
}
''' + FRAGMENT_PRODUCT_RELATION_TYPE_CONNECTION

##__________________________________________________________________||
params = [
    pytest.param(
        [
            [UPDATE_PRODUCT_RELATION_TYPE],
            {
                'variables': {
                    'typeId': 1,
                    'input': {
                        'indefArticle': "an",
                        'singular': "mmap",
                        'plural': "mmaps"
                    },
                }}
        ],
        [[QEURY], {}],
        id='update'
    ),
]

@pytest.mark.parametrize('mutation, query', params)
def test_schema_success(app, snapshot, mutation, query, mock_request_backup_db):
    assert_mutation(app, snapshot, mutation, query,
                    mock_request_backup_db, success=True)


##__________________________________________________________________||
params = [
    pytest.param(
        [
            [UPDATE_PRODUCT_RELATION_TYPE],
            {
                'variables': {
                    'typeId': 1,
                    'input': {
                        'name': "mmap",
                    },
                }}
        ],
        [[QEURY], {}],
        id='error-immutable-field'
    ),
]

@pytest.mark.parametrize('mutation, query', params)
def test_schema_error(app, snapshot, mutation, query, mock_request_backup_db):
    assert_mutation(app, snapshot, mutation, query,
                    mock_request_backup_db, success=False)

##__________________________________________________________________||



import pytest

from ..funcs import assert_mutation

from ..gql import CREATE_PRODUCT

QEURY = '''
{
  allProducts {
    edges {
      node {
        name
      }
    }
  }
  allProductRelations {
    edges {
      node {
        type_ {
          name
        }
        self_ {
          name
        }
        other {
          name
        }
      }
    }
  }
  allProductFilePaths {
    edges {
      node {
        path
      }
    }
  }
}
'''


# __________________________________________________________________||
params = [
    pytest.param(
        [
            [CREATE_PRODUCT],
            {'variables': {
                'input': {
                    'typeId': 2,
                    'name': "beam111",
                    'contact': "contact-person",
                    'dateProduced': "2020-02-20",
                    'producedBy': "producer",
                    'postedBy': "poster",
                    'note': "- Item 1",
                    'paths': [
                        "/path/to/new/product1",
                        "/another/location/of/product1"
                    ],
                    'relations' : [
                        {'typeId': 1, 'productId': 1 },
                        {'typeId': 1, 'productId': 5 }
                    ]
                }
            }},
        ],
        [[QEURY], {}],
        id='create'
    ),
    pytest.param(
        [
            [CREATE_PRODUCT],
            {'variables': {
                'input': {
                    'typeId': 1,
                    'name': "product1",
                }
            }},
        ],
        [[QEURY], {}],
        id='minimum'
    ),
    pytest.param(
        [
            [CREATE_PRODUCT],
            {'variables': {
                'input': {
                    'typeId': 2,
                    'name': "map1",
                    'contact': "contact-person",
                    'producedBy': "pwg-pmn",
                    'paths': [
                        "/path/to/new/product1",
                        "/another/location/of/product1"
                    ]
                }
            }},
        ],
        [[QEURY], {}],
        id='the-same-name-different-type'
    ),
]

@pytest.mark.parametrize('mutation, query', params)
def test_schema_success(app, snapshot, mutation, query, mock_request_backup_db):
    assert_mutation(app, snapshot, mutation, query,
                    mock_request_backup_db, success=True)

# __________________________________________________________________||
params = [
    pytest.param(
        [
            [CREATE_PRODUCT],
            {'variables': {
                'input': {
                    'typeId': 1,
                }
            }},
        ],
        [[QEURY], {}],
        id='error-no-name'
    ),
    pytest.param(
        [
            [CREATE_PRODUCT],
            {'variables': {
                'input': {
                    'typeId': 1,
                    'name': "map1",
                }
            }},
        ],
        [[QEURY], {}],
        id='error-the-same-type-and-name'
    ),
]

@pytest.mark.parametrize('mutation, query', params)
def test_schema_error(app, snapshot, mutation, query, mock_request_backup_db):
    assert_mutation(app, snapshot, mutation, query,
                    mock_request_backup_db, success=False)

# __________________________________________________________________||



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots['test_disable_mutation[kwargs0] 1'] = {
    'data': {
        'Mutation': {
            'description': None,
            'fields': [
                {
                    'name': 'createProduct'
                },
                {
                    'name': 'updateProduct'
                },
                {
                    'name': 'deleteProduct'
                },
                {
                    'name': 'createProductFilePath'
                },
                {
                    'name': 'updateProductFilePath'
                },
                {
                    'name': 'deleteProductFilePath'
                },
                {
                    'name': 'createProductType'
                },
                {
                    'name': 'updateProductType'
                },
                {
                    'name': 'deleteProductType'
                },
                {
                    'name': 'createProductRelationTypes'
                },
                {
                    'name': 'updateProductRelationType'
                },
                {
                    'name': 'deleteProductRelationTypes'
                },
                {
                    'name': 'createProductRelation'
                },
                {
                    'name': 'deleteProductRelation'
                },
                {
                    'name': 'githubAuth'
                },
                {
                    'name': 'storeAdminAppToken'
                }
            ],
            'name': 'Mutation'
        },
        'product': {
            'description': None,
            'fields': [
                {
                    'name': 'version'
                },
                {
                    'name': 'node'
                },
                {
                    'name': 'webConfig'
                },
                {
                    'name': 'allProductTypes'
                },
                {
                    'name': 'productType'
                },
                {
                    'name': 'allProducts'
                },
                {
                    'name': 'allProductFilePaths'
                },
                {
                    'name': 'product'
                },
                {
                    'name': 'allProductRelationTypes'
                },
                {
                    'name': 'productRelationType'
                },
                {
                    'name': 'allProductRelations'
                },
                {
                    'name': 'productRelation'
                },
                {
                    'name': 'githubUser'
                },
                {
                    'name': 'oauthAppInfo'
                }
            ],
            'name': 'Query'
        }
    }
}

snapshots['test_disable_mutation[kwargs1] 1'] = {
    'data': {
        'Mutation': None,
        'product': {
            'description': None,
            'fields': [
                {
                    'name': 'version'
                },
                {
                    'name': 'node'
                },
                {
                    'name': 'webConfig'
                },
                {
                    'name': 'allProductTypes'
                },
                {
                    'name': 'productType'
                },
                {
                    'name': 'allProducts'
                },
                {
                    'name': 'allProductFilePaths'
                },
                {
                    'name': 'product'
                },
                {
                    'name': 'allProductRelationTypes'
                },
                {
                    'name': 'productRelationType'
                },
                {
                    'name': 'allProductRelations'
                },
                {
                    'name': 'productRelation'
                },
                {
                    'name': 'githubUser'
                },
                {
                    'name': 'oauthAppInfo'
                }
            ],
            'name': 'Query'
        }
    }
}

snapshots['test_disable_mutation[kwargs2] 1'] = {
    'data': {
        'Mutation': {
            'description': None,
            'fields': [
                {
                    'name': 'createProduct'
                },
                {
                    'name': 'updateProduct'
                },
                {
                    'name': 'deleteProduct'
                },
                {
                    'name': 'createProductFilePath'
                },
                {
                    'name': 'updateProductFilePath'
                },
                {
                    'name': 'deleteProductFilePath'
                },
                {
                    'name': 'createProductType'
                },
                {
                    'name': 'updateProductType'
                },
                {
                    'name': 'deleteProductType'
                },
                {
                    'name': 'createProductRelationTypes'
                },
                {
                    'name': 'updateProductRelationType'
                },
                {
                    'name': 'deleteProductRelationTypes'
                },
                {
                    'name': 'createProductRelation'
                },
                {
                    'name': 'deleteProductRelation'
                },
                {
                    'name': 'githubAuth'
                },
                {
                    'name': 'storeAdminAppToken'
                }
            ],
            'name': 'Mutation'
        },
        'product': {
            'description': None,
            'fields': [
                {
                    'name': 'version'
                },
                {
                    'name': 'node'
                },
                {
                    'name': 'webConfig'
                },
                {
                    'name': 'allProductTypes'
                },
                {
                    'name': 'productType'
                },
                {
                    'name': 'allProducts'
                },
                {
                    'name': 'allProductFilePaths'
                },
                {
                    'name': 'product'
                },
                {
                    'name': 'allProductRelationTypes'
                },
                {
                    'name': 'productRelationType'
                },
                {
                    'name': 'allProductRelations'
                },
                {
                    'name': 'productRelation'
                },
                {
                    'name': 'githubUser'
                },
                {
                    'name': 'oauthAppInfo'
                }
            ],
            'name': 'Query'
        }
    }
}




import pytest
import unittest.mock as mock

from acondbs.misc import githubauth

##__________________________________________________________________||
@pytest.fixture(autouse=True)
def mock_requests(monkeypatch):
    y = mock.Mock()
    monkeypatch.setattr("acondbs.misc.githubauth.requests", y)
    yield y


##__________________________________________________________________||
def test_success(app, mock_requests):

    token = 'token-xxx'

    viewer = {
        "login": "octocat",
        "name": "monalisa octocat",
        "avatar_url": "https://github.com/images/error/octocat_happy.gif"
    }
    r = {'data': {'viewer': viewer}}

    mock_requests.post().json.return_value = r
    with app.app_context():
        user = githubauth.get_user(token)

    assert viewer == user

##__________________________________________________________________||
def test_error(app, mock_requests):

    token = 'token-xxx'

    r = {'message': 'Bad credentials', 'documentation_url': 'https://docs.github.com/graphql'}
    mock_requests.post().json.return_value = r
    with app.app_context():
        user = githubauth.get_user(token)

    assert user is None

##__________________________________________________________________||



import datetime
import graphene
from graphene import relay
from graphene_sqlalchemy import SQLAlchemyObjectType

from ..models import Product as ProductModel
from ..models import ProductFilePath as ProductFilePathModel
from ..models import ProductRelation as ProductRelationModel
from ..models import ProductRelationType as ProductRelationTypeModel

from ..db.sa import sa
from ..db.backup import request_backup_db

from .connection import CountedConnection
from .filter_ import PFilterableConnectionField

##__________________________________________________________________||
class Product(SQLAlchemyObjectType):
    class Meta:
        model = ProductModel
        interfaces = (relay.Node, )
        connection_class = CountedConnection
        connection_field_factory = PFilterableConnectionField.factory

# class ProductConnection(relay.Connection):
#     class Meta:
#         node = Product

## Product.connection is used instead
## https://github.com/graphql-python/graphene-sqlalchemy/issues/153#issuecomment-478744077

##__________________________________________________________________||
class RelationInputFields(graphene.InputObjectType):
    '''A relation to another product'''
    product_id = graphene.Int(required=True, description='The product ID of the other product')
    type_id = graphene.Int(required=True, description='The relation type ID')

class CommonInputFields:
    """Common input fields of mutations for creating and updating products

    """
    contact = graphene.String(
        description=('A person or group that can be contacted for questions or '
                     'issues about the product.'))
    note = graphene.String(description='Note about the product in MarkDown.')
    paths = graphene.List(
        graphene.String,
        description="Paths to the products. e.g., nersc:/go/to/my/product_v3")
    relations = graphene.InputField(
        graphene.List(RelationInputFields),
        description=('Relations to other products'))

class CreateProductInput(graphene.InputObjectType, CommonInputFields):
    '''Input to createProduct()'''
    type_id = graphene.Int(required=True, description='The product type ID')
    name = graphene.String(required=True, description='The name of the product')
    date_produced = graphene.Date(description='The date on which the product was produced')
    produced_by = graphene.String(description='The person or group that produced the product')
    posted_by = graphene.String(description='The person who entered the DB entry.')

class UpdateProductInput(graphene.InputObjectType, CommonInputFields):
    '''Input to updateProduct()'''
    updated_by = graphene.String(description='The person who updated the DB entry.')

##__________________________________________________________________||
class CreateProduct(graphene.Mutation):
    '''Create a product'''
    class Arguments:
        input = CreateProductInput(
            required=True,
            description=('the input to createProduct()')
        )

    ok = graphene.Boolean()
    product = graphene.Field(lambda: Product)

    def mutate(root, info, input):
        paths = [ProductFilePathModel(path=p) for p in input.pop('paths', [])]
        with sa.session.no_autoflush:
            relations = [
                ProductRelationModel(
                    type_=ProductRelationTypeModel.query.filter_by(type_id=r['type_id']).one(),
                    other=ProductModel.query.filter_by(product_id=r['product_id']).one()
                )
                for r in input.pop('relations', [])
            ]
        model = ProductModel(
            date_posted=datetime.date.today(),
            paths=paths,
            relations=relations,
            **input
        )
        sa.session.add(model)
        sa.session.commit()
        ok = True
        request_backup_db()
        return CreateProduct(product=model, ok=ok)

class UpdateProduct(graphene.Mutation):
    '''Update a product. Note: This is to update the DB entry about a product. If the
    product itself has been updated, a new entry should be added by
    createProduct()

    '''
    class Arguments:
        product_id = graphene.Int(
            required=True,
            description='The productId of a product to be updated.')
        input = UpdateProductInput(
            required=True,
            description='an input to updateProduct()'
        )

    ok = graphene.Boolean()
    product = graphene.Field(lambda: Product)

    def mutate(root, info, product_id, input):
        model = ProductModel.query.filter_by(product_id=product_id).one()

        # update paths
        input_paths = input.pop('paths', None)
        if input_paths is not None:
            pdict = {p.path: p for p in model.paths}
            model.paths = [
                pdict[p] if p in pdict else ProductFilePathModel(path=p)
                for p in input_paths]

        # update relations
        input_relations = input.pop('relations', None)
        if input_relations is not None:
            with sa.session.no_autoflush:
                old_relations_dict = {(r.type_id, r.self_product_id, r.other_product_id): r for r in model.relations}
                for r in input_relations:
                    rmodel = old_relations_dict.pop((r['type_id'], model.product_id, r['product_id']), None)
                    if not rmodel:
                        type_ = ProductRelationTypeModel.query.filter_by(type_id=r['type_id']).one()
                        other = ProductModel.query.filter_by(product_id=r['product_id']).one()
                        m = ProductRelationModel(self_=model, type_=type_, other=other)
                        sa.session.add(m)
                for m in old_relations_dict.values():
                    sa.session.delete(m)

        # update scalar fields
        for k, v in input.items():
            setattr(model, k, v)

        model.date_updated = datetime.date.today()

        sa.session.commit()
        ok = True
        request_backup_db()
        return UpdateProduct(product=model, ok=ok)

class DeleteProduct(graphene.Mutation):
    '''Delete a product'''
    class Arguments:
        product_id = graphene.Int(
            required=True,
            description='The productId of a product to be deleted.')

    ok = graphene.Boolean()

    def mutate(root, info, product_id):
        model = ProductModel.query.filter_by(product_id=product_id).one()
        sa.session.delete(model)
        sa.session.commit()
        ok = True
        request_backup_db()
        return DeleteProduct(ok=ok)

##__________________________________________________________________||



from flask import current_app
import graphene
from graphene import relay
from graphene_sqlalchemy import SQLAlchemyConnectionField
from graphql import GraphQLError

from .product import Product, ProductModel
from .product_file_path import ProductFilePath, ProductFilePathModel
from .product_type import ProductType, ProductTypeModel
from .product_relation_type import ProductRelationType, ProductRelationTypeModel
from .product_relation import ProductRelation, ProductRelationModel

from .web_config import WebConfig

from .filter_ import PFilterableConnectionField

from .auth import OAuthAppInfo, GitHubUser

from ..misc import githubauth

##__________________________________________________________________||
class Query(graphene.ObjectType):

    version = graphene.String()

    def resolve_version(self, info):
        from .. import __version__
        return __version__

    node = relay.Node.Field()

    web_config = graphene.Field(WebConfig)

    def resolve_web_config(self, info, **kwargs):
        return WebConfig.get_query(info).one_or_none()

    all_product_types = PFilterableConnectionField(ProductType.connection)

    product_type = graphene.Field(ProductType, type_id=graphene.Int(), name=graphene.String())

    def resolve_product_type(self, info, **kwargs):
        filter = [getattr(ProductTypeModel, k)==v for k, v in kwargs.items()]
        # e.g., [ProductTypeModel.type_id == 1, ProductTypeModel.name == 'map']

        return ProductType.get_query(info).filter(*filter).one_or_none()

    all_products = PFilterableConnectionField(Product.connection)
    all_product_file_paths = PFilterableConnectionField(ProductFilePath.connection)

    product = graphene.Field(
        Product,
        product_id=graphene.Int(),
        type_id=graphene.Int(),
        name=graphene.String())

    def resolve_product(self, info, **kwargs):

        filter = [getattr(ProductModel, k)==v for k, v in kwargs.items()]
        # e.g., [ProductModel.type_id == 1, ProductModel.name == 'map_001']

        return Product.get_query(info).filter(*filter).one_or_none()

    all_product_relation_types = PFilterableConnectionField(ProductRelationType.connection)

    product_relation_type = graphene.Field(ProductRelationType, type_id=graphene.Int(), name=graphene.String())

    def resolve_product_relation_type(self, info, **kwargs):
        filter = [getattr(ProductRelationTypeModel, k)==v for k, v in kwargs.items()]
        return ProductRelationType.get_query(info).filter(*filter).one_or_none()

    all_product_relations = PFilterableConnectionField(ProductRelation.connection)

    product_relation = graphene.Field(ProductRelation, relation_id=graphene.Int())

    def resolve_product_relation(self, info, **kwargs):
        filter = [getattr(ProductRelationModel, k)==v for k, v in kwargs.items()]
        return ProductRelation.get_query(info).filter(*filter).one_or_none()

    github_user = graphene.Field(GitHubUser)

    def resolve_github_user(self, info):

        auth = info.context.headers.get('Authorization')
        # e.g., 'Bearer "xxxx"'

        if not auth:
            raise GraphQLError('Authorization is required')

        token = auth.split()[1].strip('"')
        # e.g., "xxxx"

        user = githubauth.get_user(token)
        if not user:
            raise GraphQLError('Unsuccessful to obtain the user')

        return GitHubUser(**user);

    oauth_app_info = graphene.Field(
        OAuthAppInfo,
        admin=graphene.Boolean(default_value=False)
        )

    def resolve_oauth_app_info(self, info, admin):
        print(admin)
        if admin:
            return OAuthAppInfo(
                client_id=current_app.config['GITHUB_AUTH_ADMIN_CLIENT_ID'],
                authorize_url=current_app.config['GITHUB_AUTH_AUTHORIZE_URL'],
                token_url=current_app.config['GITHUB_AUTH_TOKEN_URL'],
                redirect_uri=current_app.config['GITHUB_AUTH_ADMIN_REDIRECT_URI']
            )

        return OAuthAppInfo(
            client_id=current_app.config['GITHUB_AUTH_CLIENT_ID'],
            authorize_url=current_app.config['GITHUB_AUTH_AUTHORIZE_URL'],
            token_url=current_app.config['GITHUB_AUTH_TOKEN_URL'],
            redirect_uri=current_app.config['GITHUB_AUTH_REDIRECT_URI']
        )
##__________________________________________________________________||



from .product import (
    FRAGMENT_PRODUCT_SHALLOW,
    FRAGMENT_PRODUCT_CONNECTION_SHALLOW,
    FRAGMENT_PRODUCT_DEEP,
    FRAGMENT_PRODUCT_CONNECTION_DEEP,
    CREATE_PRODUCT,
    DELETE_PRODUCT,
    UPDATE_PRODUCT
)

from .product_relation_type import (
    FRAGMENT_PRODUCT_RELATION_TYPE,
    FRAGMENT_PRODUCT_RELATION_TYPE_CONNECTION,
    CREATE_PRODUCT_RELATION_TYPES,
    UPDATE_PRODUCT_RELATION_TYPE,
    DELETE_PRODUCT_RELATION_TYPES
    )

from .product_relation import (
    FRAGMENT_PRODUCT_RELATION,
    FRAGMENT_PRODUCT_RELATION_CONNECTION,
    CREATE_PRODUCT_RELATION,
    DELETE_PRODUCT_RELATION
    )

from .product_type import (
    FRAGMENT_PRODUCT_TYPE,
    FRAGMENT_PRODUCT_TYPE_CONNECTION,
    CREATE_PRODUCT_TYPE,
    UPDATE_PRODUCT_TYPE,
    DELETE_PRODUCT_TYPE
    )



import pytest

from acondbs.db.sa import sa
from acondbs.models import ProductType, Product

# These tests are written primarily for the developer to understand
# how models in flask_sqlalchemy work.

# __________________________________________________________________||
@pytest.fixture
def app(app):
    y = app

    type_map = ProductType(name='map')
    map1 = Product(name="map1", type_=type_map)
    map2 = Product(name="map2", type_=type_map)
    map3 = Product(name="map3", type_=type_map)

    with y.app_context():
        sa.session.add(type_map)
        sa.session.commit()

    yield y

# __________________________________________________________________||
def test_simple(app):
    '''A simple test of deleting an object
    '''

    with app.app_context():

        assert 3 == len(Product.query.all())

    with app.app_context():
        product1 = Product.query.filter_by(name='map1').first()
        sa.session.delete(product1)
        sa.session.commit()

    with app.app_context():

        assert 2 == len(Product.query.all())

        # the product is no longer found
        product1 = Product.query.filter_by(name='map1').first()
        assert product1 is None

# __________________________________________________________________||




##__________________________________________________________________||
FRAGMENT_PRODUCT_TYPE = '''
fragment fragmentProductType on ProductType {
  typeId
  name
  order
  indefArticle
  singular
  plural
  icon
  products {
    edges {
      node {
        name
      }
    }
  }
}
'''

FRAGMENT_PRODUCT_TYPE_CONNECTION = '''
fragment fragmentProductTypeConnection on ProductTypeConnection {
  edges {
    node {
      ...fragmentProductType
    }
  }
}
''' + FRAGMENT_PRODUCT_TYPE

##__________________________________________________________________||
CREATE_PRODUCT_TYPE = '''
mutation CreateProductType($input: CreateProductTypeInput!) {
  createProductType(input: $input) {
    ok
    productType {
      ...fragmentProductType
    }
  }
}
''' + FRAGMENT_PRODUCT_TYPE

DELETE_PRODUCT_TYPE = '''
mutation DeleteProductType($typeId: Int!) {
  deleteProductType(typeId: $typeId) {
    ok
  }
}
'''

UPDATE_PRODUCT_TYPE = '''
mutation UpdateProductType($typeId: Int!, $input: UpdateProductTypeInput!) {
  updateProductType(typeId: $typeId,input: $input) {
    ok
    productType {
      ...fragmentProductType
    }
  }
}
''' + FRAGMENT_PRODUCT_TYPE

##__________________________________________________________________||



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import GenericRepr, Snapshot


snapshots = Snapshot()

snapshots['test_non_empty 1'] = {
    'admin_app_token': [
        {
            'token': b'aKjGknYDHY39Z2xAaN7+sQ==',
            'token_id': 1
        }
    ],
    'product_file_paths': [
        {
            'note': '',
            'path': 'nersc:/go/to/my/maps',
            'path_id': 1,
            'product_id': 1001
        },
        {
            'note': 'lat only',
            'path': 'nersc:/go/to/my/maps_v2',
            'path_id': 2,
            'product_id': 1012
        },
        {
            'note': 'lat only',
            'path': 'abcde:/path/to/the/maps_v2',
            'path_id': 3,
            'product_id': 1012
        },
        {
            'note': 'lat only',
            'path': 'nersc:/go/to/my/maps_v3',
            'path_id': 4,
            'product_id': 1013
        },
        {
            'note': '',
            'path': 'BEAM_DEPOT/Beams/20190304',
            'path_id': 5,
            'product_id': 1070
        },
        {
            'note': '',
            'path': 'BEAM_DEPOT/Beams/20190607',
            'path_id': 6,
            'product_id': 1120
        },
        {
            'note': '',
            'path': 'BEAM_DEPOT/Beams/20200123',
            'path_id': 7,
            'product_id': 1130
        },
        {
            'note': '',
            'path': 'BEAM_DEPOT/Beams/20200207',
            'path_id': 8,
            'product_id': 1150
        },
        {
            'note': '',
            'path': 'nersc:/go/to/my/simulations',
            'path_id': 9,
            'product_id': 1002
        },
        {
            'note': '',
            'path': 'abcde:/path/to/the/simulations',
            'path_id': 10,
            'product_id': 1002
        }
    ],
    'product_relation_types': [
        {
            'indef_article': 'a',
            'name': 'parent',
            'plural': 'parents',
            'reverse_type_id': 2,
            'singular': 'parent',
            'type_id': 1
        },
        {
            'indef_article': 'a',
            'name': 'child',
            'plural': 'children',
            'reverse_type_id': 1,
            'singular': 'child',
            'type_id': 2
        },
        {
            'indef_article': 'a',
            'name': 'plaintiff',
            'plural': 'plaintiffs',
            'reverse_type_id': 4,
            'singular': 'plaintiff',
            'type_id': 3
        },
        {
            'indef_article': 'a',
            'name': 'defendant',
            'plural': 'defendants',
            'reverse_type_id': 3,
            'singular': 'defendant',
            'type_id': 4
        }
    ],
    'product_relations': [
        {
            'other_product_id': 1012,
            'relation_id': 1,
            'reverse_relation_id': 2,
            'self_product_id': 1130,
            'type_id': 1
        },
        {
            'other_product_id': 1130,
            'relation_id': 2,
            'reverse_relation_id': 1,
            'self_product_id': 1012,
            'type_id': 2
        },
        {
            'other_product_id': 1013,
            'relation_id': 3,
            'reverse_relation_id': 4,
            'self_product_id': 1150,
            'type_id': 1
        },
        {
            'other_product_id': 1150,
            'relation_id': 4,
            'reverse_relation_id': 3,
            'self_product_id': 1013,
            'type_id': 2
        },
        {
            'other_product_id': 1130,
            'relation_id': 5,
            'reverse_relation_id': 6,
            'self_product_id': 1150,
            'type_id': 1
        },
        {
            'other_product_id': 1150,
            'relation_id': 6,
            'reverse_relation_id': 5,
            'self_product_id': 1130,
            'type_id': 2
        }
    ],
    'product_types': [
        {
            'icon': 'mdi-map',
            'indef_article': 'a',
            'name': 'map',
            'order': 2,
            'plural': 'maps',
            'singular': 'map',
            'type_id': 1
        },
        {
            'icon': 'mdi-spotlight-beam',
            'indef_article': 'a',
            'name': 'beam',
            'order': 3,
            'plural': 'beams',
            'singular': 'beam',
            'type_id': 2
        },
        {
            'icon': 'mdi-creation',
            'indef_article': 'a',
            'name': 'simulation',
            'order': 1,
            'plural': 'simulations',
            'singular': 'simulation',
            'type_id': 3
        },
        {
            'icon': '',
            'indef_article': '',
            'name': 'anchor',
            'order': 4,
            'plural': '',
            'singular': '',
            'type_id': 4
        }
    ],
    'products': [
        {
            'contact': 'pwg-pmn',
            'date_posted': GenericRepr('datetime.date(2019, 2, 13)'),
            'date_produced': GenericRepr('datetime.date(2019, 2, 13)'),
            'date_updated': GenericRepr('datetime.date(2019, 2, 13)'),
            'name': 'lat20190213',
            'note': '''- This is a dummy test with a lat map
- This should not depend on any beam''',
            'posted_by': 'pwg-pmn',
            'produced_by': 'pwg-pmn',
            'product_id': 1001,
            'type_id': 1,
            'updated_by': 'pwg-pmn'
        },
        {
            'contact': 'abc-def',
            'date_posted': GenericRepr('datetime.date(2019, 3, 15)'),
            'date_produced': GenericRepr('datetime.date(2019, 3, 15)'),
            'date_updated': None,
            'name': 'xyz-s1234-20200101',
            'note': '''- note 1
- note 2''',
            'posted_by': 'abc-def',
            'produced_by': 'abc-def',
            'product_id': 1002,
            'type_id': 3,
            'updated_by': ''
        },
        {
            'contact': 'pwg-pmn',
            'date_posted': GenericRepr('datetime.date(2018, 1, 1)'),
            'date_produced': GenericRepr('datetime.date(2018, 1, 1)'),
            'date_updated': None,
            'name': '20180101',
            'note': '- test entry',
            'posted_by': 'pwg-pmn',
            'produced_by': 'pwg-pmn',
            'product_id': 1010,
            'type_id': 2,
            'updated_by': ''
        },
        {
            'contact': 'pwg-pmn',
            'date_posted': GenericRepr('datetime.date(2020, 1, 20)'),
            'date_produced': GenericRepr('datetime.date(2020, 1, 20)'),
            'date_updated': GenericRepr('datetime.date(2020, 1, 20)'),
            'name': 'lat20200120',
            'note': '''- This is a dummy test with a lat map
- A beam depends on this map''',
            'posted_by': 'pwg-pmn',
            'produced_by': 'pwg-pmn',
            'product_id': 1012,
            'type_id': 1,
            'updated_by': 'pwg-pmn'
        },
        {
            'contact': 'pwg-pmn',
            'date_posted': GenericRepr('datetime.date(2020, 2, 1)'),
            'date_produced': GenericRepr('datetime.date(2020, 2, 1)'),
            'date_updated': GenericRepr('datetime.date(2020, 2, 1)'),
            'name': 'lat20200201',
            'note': '''- This is a dummy test with a lat map
- A beam depends on this map''',
            'posted_by': 'pwg-pmn',
            'produced_by': 'pwg-pmn',
            'product_id': 1013,
            'type_id': 1,
            'updated_by': 'pwg-pmn'
        },
        {
            'contact': 'pwg-pmn',
            'date_posted': GenericRepr('datetime.date(2019, 3, 4)'),
            'date_produced': GenericRepr('datetime.date(2019, 3, 4)'),
            'date_updated': None,
            'name': '20190304',
            'note': '- test entry',
            'posted_by': 'pwg-pmn',
            'produced_by': 'pwg-pmn',
            'product_id': 1070,
            'type_id': 2,
            'updated_by': ''
        },
        {
            'contact': 'pwg-pmn',
            'date_posted': GenericRepr('datetime.date(2019, 6, 7)'),
            'date_produced': GenericRepr('datetime.date(2019, 6, 7)'),
            'date_updated': None,
            'name': '20190607',
            'note': '- test entry',
            'posted_by': 'pwg-pmn',
            'produced_by': 'pwg-pmn',
            'product_id': 1120,
            'type_id': 2,
            'updated_by': ''
        },
        {
            'contact': 'pwg-pmn',
            'date_posted': GenericRepr('datetime.date(2020, 1, 23)'),
            'date_produced': GenericRepr('datetime.date(2020, 1, 23)'),
            'date_updated': None,
            'name': '20200123',
            'note': '- test entry',
            'posted_by': 'pwg-pmn',
            'produced_by': 'pwg-pmn',
            'product_id': 1130,
            'type_id': 2,
            'updated_by': ''
        },
        {
            'contact': 'pwg-pmn',
            'date_posted': GenericRepr('datetime.date(2020, 2, 7)'),
            'date_produced': GenericRepr('datetime.date(2020, 2, 7)'),
            'date_updated': None,
            'name': '20200207',
            'note': '- test entry',
            'posted_by': 'pwg-pmn',
            'produced_by': 'pwg-pmn',
            'product_id': 1150,
            'type_id': 2,
            'updated_by': ''
        }
    ],
    'web_config': [
        {
            'config_id': 1,
            'devtool_loadingstate': True,
            'head_title': 'Product DB (test)',
            'product_creation_dialog': True,
            'product_deletion_dialog': False,
            'product_update_dialog': True,
            'toolbar_title': 'Product DB (test)'
        }
    ]
}



import pytest

from acondbs import create_app
from acondbs.db.ops import define_tables

from acondbs.db.sa import sa
from acondbs.models import (
    WebConfig
    )

##__________________________________________________________________||
@pytest.fixture
def app_empty():
    database_uri ="sqlite:///:memory:"
    y = create_app(SQLALCHEMY_DATABASE_URI=database_uri)
    with y.app_context():
        define_tables()
    yield y

@pytest.fixture
def app(app_empty):

    y = app_empty

    c = WebConfig(
        config_id=1,
        head_title="Head Title",
        toolbar_title="Toolbar Title",
        devtool_loadingstate=True,
        product_creation_dialog=False,
        product_update_dialog=True,
        product_deletion_dialog=True
    )

    with y.app_context():
        sa.session.add(c)
        sa.session.commit()
    yield y

##__________________________________________________________________||



import pytest
import textwrap

from ..funcs import assert_query

from ..gql import FRAGMENT_PRODUCT_RELATION_TYPE

##__________________________________________________________________||
params = [
    pytest.param(
        [textwrap.dedent('''
          {
            productRelationType(typeId: 1) {
              ...fragmentProductRelationType
            }
          }
        ''') + FRAGMENT_PRODUCT_RELATION_TYPE,],
        {},
        id='type_id'
    ),
    pytest.param(
        [textwrap.dedent('''
          {
            productRelationType(name: "parent") {
              ...fragmentProductRelationType
            }
          }
        ''') + FRAGMENT_PRODUCT_RELATION_TYPE,],
        {},
        id='name'
    ),
    pytest.param(
        [textwrap.dedent('''
          {
            productRelationType(typeId: 1, name: "parent") {
              ...fragmentProductRelationType
            }
          }
        ''') + FRAGMENT_PRODUCT_RELATION_TYPE,],
        {},
        id='type_id-and-name'
    ),
    pytest.param(
        [textwrap.dedent('''
          {
            productRelationType(typeId: 2, name: "parent") {
              ...fragmentProductRelationType
            }
          }
        ''') + FRAGMENT_PRODUCT_RELATION_TYPE,],
        {},
        id='type_id-and-name-nonexistent)'
    ),
]

##__________________________________________________________________||
@pytest.mark.parametrize('args, kwags', params)
def test_schema(app, snapshot, args, kwags):
    assert_query(app, snapshot, [args, kwags])

# __________________________________________________________________||



from ..db.sa import sa

from sqlalchemy.event import listens_for

##__________________________________________________________________||
class ProductRelationType(sa.Model):
    __tablename__ = 'product_relation_types'
    type_id = sa.Column(sa.Integer(), primary_key=True)
    name = sa.Column(sa.Text(), nullable=False, unique=True, index=True)
    reverse_type_id = sa.Column(sa.ForeignKey('product_relation_types.type_id'))
    reverse = sa.relationship(
        lambda: ProductRelationType,
        uselist=False,
        foreign_keys=[reverse_type_id],
        remote_side="ProductRelationType.type_id",
        cascade="all",
        post_update=True)
    indef_article = sa.Column(sa.Text())
    singular = sa.Column(sa.Text())
    plural = sa.Column(sa.Text())

##__________________________________________________________________||
@listens_for(ProductRelationType.reverse, 'set')
def set_reverse(target, value, oldvalue, initiator):
    """set the reciprocating reverse

    For example, if the reverse of the type1 is set type2,
       type1.reverse = type2
    this function sets the reverse of the type2 to type1,
       type2.reverse = type1

    """

    type1 = target
    type2 = value

    if type1 is type2:
        return

    try:
        if type1.__avoid_recursive:
            return
    except:
        pass

    if not type2.reverse:
        type2.__avoid_recursive = True
        type2.reverse = type1
        del type2.__avoid_recursive

##__________________________________________________________________||



import pytest

from sqlalchemy import exc

from acondbs.db.sa import sa
from acondbs.models import ProductType, Product

# __________________________________________________________________||
def test_product(app_empty):
    app = app_empty

    type_map = ProductType(name='map')
    map1 = Product(name="map1", type_=type_map)

    with app.app_context():
        sa.session.add(map1)
        sa.session.commit()

    with app.app_context():
        map1 = Product.query.filter_by(name='map1').one_or_none()
        assert map1 is not None
        type_map = map1.type_
        assert 'map' == type_map.name
        assert [map1] == type_map.products

# __________________________________________________________________||
def test_constraint_type_required_add(app_empty):
    app = app_empty

    map1 = Product(name="map1")

    with app.app_context():
        sa.session.add(map1)
        with pytest.raises(exc.IntegrityError):
            sa.session.commit()

    with app.app_context():
        map1 = Product.query.filter_by(name='map1').one_or_none()
        assert map1 is None

# __________________________________________________________________||
def test_constraint_type_required_delete(app_empty):
    app = app_empty

    type_map = ProductType(name='map')
    map1 = Product(name="map1", type_=type_map)

    with app.app_context():
        sa.session.add(map1)
        sa.session.commit()

    # fail to delete a type with a product
    with app.app_context():
        type_map = ProductType.query.filter_by(name='map').one_or_none()
        sa.session.delete(type_map)
        with pytest.raises(exc.IntegrityError):
            sa.session.commit()

    # assert the type and the product are still there
    with app.app_context():
        map1 = Product.query.filter_by(name='map1').one_or_none()
        assert map1 is not None
        type_map = map1.type_
        assert 'map' == type_map.name
        assert [map1] == type_map.products

    # delete the product
    with app.app_context():
        map1 = Product.query.filter_by(name='map1').one_or_none()
        sa.session.delete(map1)
        sa.session.commit()

    # assert the type still exists
    with app.app_context():
        type_map = ProductType.query.filter_by(name='map').one_or_none()
        assert type_map is not None
        assert [] == type_map.products

    # delete the type
    with app.app_context():
        type_map = ProductType.query.filter_by(name='map').one_or_none()
        sa.session.delete(type_map)
        sa.session.commit()

# __________________________________________________________________||



import pytest

from ..funcs import assert_query

OAUTH_APP_INFO = '''
query OauthAppInfo($admin: Boolean) {
  oauthAppInfo(admin: $admin) {
    clientId
    authorizeUrl
    tokenUrl
    redirectUri
  }
}
'''

##__________________________________________________________________||
params = [
    pytest.param(
        [OAUTH_APP_INFO],
        {'variables': {'admin': False}},
        id='app'
    ),
    pytest.param(
        [OAUTH_APP_INFO],
        {'variables': {'admin': True}},
        id='app-admin'
    )
]

##__________________________________________________________________||
@pytest.mark.parametrize('args, kwags', params)
def test_schema(app, snapshot, args, kwags):
    assert_query(app, snapshot, [args, kwags])

##__________________________________________________________________||



import pytest

from ..funcs import assert_mutation

from ..gql import DELETE_PRODUCT

QEURY = '''
{
  allProducts {
    edges {
      node {
        name
      }
    }
  }
  allProductRelations {
    edges {
      node {
        type_ {
          name
        }
        self_ {
          name
        }
        other {
          name
        }
      }
    }
  }
  allProductFilePaths {
    edges {
      node {
        path
      }
    }
  }
}
'''


# __________________________________________________________________||
params = [
    pytest.param(
        [
            [DELETE_PRODUCT],
            {'variables': {'productId': 1}},
        ],
        [[QEURY], {}],
        id='delete'
    ),
]


@pytest.mark.parametrize('mutation, query', params)
def test_schema_success(app, snapshot, mutation, query, mock_request_backup_db):
    assert_mutation(app, snapshot, mutation, query,
                    mock_request_backup_db, success=True)


# __________________________________________________________________||
params = [
    pytest.param(
        [
            [DELETE_PRODUCT],
            {'variables': {'productId': 512}},
        ],
        [[QEURY], {}],
        id='error'
    ),
]


@pytest.mark.parametrize('mutation, query', params)
def test_schema_error(app, snapshot, mutation, query, mock_request_backup_db):
    assert_mutation(app, snapshot, mutation, query,
                    mock_request_backup_db, success=False)

# __________________________________________________________________||



import pytest
import textwrap

from ..funcs import assert_query

from ..gql import FRAGMENT_PRODUCT_TYPE

##__________________________________________________________________||
params = [
    pytest.param(
        [textwrap.dedent('''
          {
            productType(typeId: 1) {
              ...fragmentProductType
            }
          }
         ''') + FRAGMENT_PRODUCT_TYPE,],
        {},
        id='type_id'
    ),
    pytest.param(
        [textwrap.dedent('''
          {
            productType(name: "map") {
              ...fragmentProductType
            }
          }
         ''') + FRAGMENT_PRODUCT_TYPE,],
        {},
        id='name'
    ),
    pytest.param(
        [textwrap.dedent('''
          {
            productType(typeId: 1, name: "map") {
              ...fragmentProductType
            }
          }
         ''') + FRAGMENT_PRODUCT_TYPE,],
        {},
        id='type_id-and-name'
    ),
    pytest.param(
        [textwrap.dedent('''
          {
            productType(typeId: 2, name: "map") {
              ...fragmentProductType
            }
          }
         ''') + FRAGMENT_PRODUCT_TYPE,],
        {},
        id='type_id-and-name-nonexistent'
    ),
    pytest.param(
        ['''
          {
            productType(typeId: 1) {
              typeId
              name
              order
              indefArticle
              singular
              plural
              icon
              products(sort: DATE_PRODUCED_DESC) {
                edges {
                  node {
                    name
                  }
                }
              }
            }
          }
         ''',],
        {},
        id='type_id-sort-products'
    ),
]

##__________________________________________________________________||
@pytest.mark.parametrize('args, kwags', params)
def test_schema(app, snapshot, args, kwags):
    assert_query(app, snapshot, [args, kwags])

# __________________________________________________________________||




##__________________________________________________________________||
FRAGMENT_PRODUCT_SHALLOW = '''
fragment fragmentProductShallow on Product {
  productId
  typeId
  name
}
'''

FRAGMENT_PRODUCT_CONNECTION_SHALLOW = '''
fragment fragmentProductConnectionShallow on ProductConnection {
  edges {
    node {
     ...fragmentProductShallow
    }
  }
}
''' + FRAGMENT_PRODUCT_SHALLOW

##__________________________________________________________________||
FRAGMENT_PRODUCT_DEEP = '''
fragment fragmentProductDeep on Product {
  productId
  typeId
  type_ {
    typeId
    name
  }
  name
  contact
  dateProduced
  producedBy
  datePosted
  postedBy
  dateUpdated
  updatedBy
  paths {
    edges {
      node {
        pathId
        path
        note
      }
    }
  }
  relations {
    edges {
      node {
        relationId
        typeId
        type_ {
          typeId
          name
        }
        otherProductId
        other {
          productId
          typeId
          type_ {
            typeId
            name
          }
          name
        }
        reverseRelationId
        reverse {
          relationId
          typeId
          type_ {
            typeId
            name
          }
        }
      }
    }
  }
  note
}
'''

FRAGMENT_PRODUCT_CONNECTION_DEEP = '''
fragment fragmentProductConnectionDeep on ProductConnection {
  edges {
    node {
     ...fragmentProductDeep
    }
  }
}
''' + FRAGMENT_PRODUCT_DEEP

##__________________________________________________________________||
CREATE_PRODUCT = '''
mutation CreateProduct($input: CreateProductInput!) {
  createProduct(input: $input) {
    ok
    product {
      ...fragmentProductDeep
    }
  }
}
''' + FRAGMENT_PRODUCT_DEEP

DELETE_PRODUCT = '''
mutation DeleteProduct($productId: Int!) {
  deleteProduct(productId: $productId) {
    ok
  }
}
'''

UPDATE_PRODUCT = '''
mutation($productId: Int!, $input: UpdateProductInput!) {
  updateProduct(productId: $productId, input: $input) {
    ok
    product {
      ...fragmentProductDeep
    }
  }
}
''' + FRAGMENT_PRODUCT_DEEP

##__________________________________________________________________||



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots['test_schema_success[update] 1'] = {
    'data': {
        'updateProductType': {
            'ok': True,
            'productType': {
                'icon': 'mdi-compass',
                'indefArticle': 'a',
                'name': 'map',
                'order': 5,
                'plural': 'compasses',
                'products': {
                    'edges': [
                        {
                            'node': {
                                'name': 'map1'
                            }
                        },
                        {
                            'node': {
                                'name': 'map2'
                            }
                        },
                        {
                            'node': {
                                'name': 'map3'
                            }
                        }
                    ]
                },
                'singular': 'compass',
                'typeId': '1'
            }
        }
    }
}

snapshots['test_schema_success[update] 2'] = {
    'data': {
        'allProductTypes': {
            'edges': [
                {
                    'node': {
                        'name': 'map'
                    }
                },
                {
                    'node': {
                        'name': 'beam'
                    }
                }
            ]
        }
    }
}

snapshots['test_schema_error[error-nonexistent] 1'] = {
    'data': {
        'allProductTypes': {
            'edges': [
                {
                    'node': {
                        'name': 'map'
                    }
                },
                {
                    'node': {
                        'name': 'beam'
                    }
                }
            ]
        }
    }
}
